#if defined(__APPLE__)
/* We must not leave undefined reference on MacOSX  */
/* Plus, we cannot use dlfcn in initialization time */
# include <mach-o/dyld.h>
#endif

#ifdef USE_DLFCN
# define _GNU_SOURCE
# include <dlfcn.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <assert.h>

#include <sys/mman.h>

#define PAGESIZE 4096

#if defined(__i386__)
static const char jmp_code[] =
    "\xb8\0\0\0\0"  // mov EAX, addr
    "\xff\xe0"      // jmp EAX
    ;
static const int addr_offset = 1;
#elif defined(__x86_64__)
static const char jmp_code[] =
    "\x48\xb8\0\0\0\0\0\0\0\0"  // mov RAX, addr
    "\xff\xe0"                  // jmp RAX
    ;
static const int addr_offset = 2;
#else
# error "Unsupported architecture!"
#endif

#define jmp_code_size (sizeof(jmp_code) - 1)

static char orig_code[jmp_code_size];

static int (*real_main)(int argc, char** argv);

static char* align(char* m) {
    return (char*)((((uintptr_t)m+PAGESIZE-1)&~(PAGESIZE-1))-PAGESIZE);
}

static char* mprotect_main(int prot) {
    char* m = (char*)real_main;
    char* aligned = align(m);
    char* aligned2 = align(m + jmp_code_size);
    int size = aligned2 - aligned + PAGESIZE;
    //printf("%p => %p+%x\n", m, aligned, size);
    if (mprotect(aligned, size, prot) < 0) {
        perror("main_hook: mprotect");
        exit(1);
    }
    return m;
}

static void init_main_hook(int (*main_hook)(int, char**)) {
#if defined(__APPLE__)
    NSModule nsmodule;
    void* fp = NULL;
    _dyld_lookup_and_bind_fully("_main", &fp, &nsmodule);
    real_main = (int (*)(int, char**))fp;
#elif defined(USE_DLFCN)
    real_main = (int (*)(int, char**))dlsym(RTLD_DEFAULT, "main");
#else
    int main(int argc, char** argv);
    real_main = &main;
#endif
    assert(real_main);

    char* m = mprotect_main(PROT_READ | PROT_EXEC | PROT_WRITE);

    memcpy(orig_code, m, jmp_code_size);
    memcpy(m, jmp_code, jmp_code_size);
    void** addrp = (void**)(m + addr_offset);
    *addrp = (void*)main_hook;

    mprotect_main(PROT_READ | PROT_EXEC);
}

static int quit_main_hook(int argc, char** argv) {
    char* m = mprotect_main(PROT_READ | PROT_EXEC | PROT_WRITE);

    memcpy(m, orig_code, jmp_code_size);

    mprotect_main(PROT_READ | PROT_EXEC);

    return real_main(argc, argv);
}
